@@ -12,6 +12,16 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 # ==============================================================================
+# autopep8: off
+
+# constant
+from enum import Enum
+
+class phase(Enum):
+  train = 0
+  val = 1
+  test = 2
+  viz = 3
 
 # 1 - independent
 from .utils.logger import logger
@@ -23,8 +33,14 @@ from .utils import drawer
 from .utils import parser
 from .utils import export
 
+# media
+from .media import ply
+from .media import media
+# from .media import raw
+from .media import flow
+from .media import mesh
+
 # 2 - utils level
-from .utils import media
 from .utils import checkpoint
 from .utils import runner
 from . import evaluator
@@ -45,11 +61,7 @@ from .utils import flops
 # extra repo
 from . import contrib
 
-# 7 - constant
-from enum import Enum
+# optim
+from . import optim
 
-class phase(Enum):
-  train = 0
-  val = 1
-  test = 2
-  viz = 3
+# autopep8: on
