@@ -12,7 +12,4 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 # ==============================================================================
-"""Reference
-  https://github.com/cunjian/pytorch_face_landmark
-"""
-
+from .mobileface import MobileFace
