@@ -23,10 +23,19 @@ from .classification import ConfusionMatrixEvaluator
 from .image import ImageSimilarityEvaluator
 
 from .regression import RegressionEvaluator
-from .regression import Avec2014Evaluator
+from .avec2014 import Avec2014Evaluator
+from .quality_assess import QualityAssessEvaluator
 
 from .segmentation import SegmentationEvaluator
 from .segmentation import SaliencyEvaluator
 from .segmentation import MattingEvaluator
+from .segmentation import PointCloudSegmentEvaluator
 
 from .verification import VerificationEvaluator
+
+from .detection import CocoEvaluator
+from .widerface import WiderfaceEvaluator
+
+from .landmark import FaceLandmarkEvaluator
+
+from .optical_flow import OpticalFlowEvaluator
\ No newline at end of file
