@@ -12,11 +12,11 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 # ==============================================================================
-r"""activation
+"""activation
 """
-
 import torch
 from torch import nn
+from torch.nn import functional as F
 
 
 class SwishImplementation(torch.autograd.Function):
@@ -41,3 +41,26 @@ class MemoryEfficientSwish(nn.Module):
 
 
 Swish = MemoryEfficientSwish
+
+
+def mish(x, inplace: bool = False):
+  """Mish: A Self Regularized Non-Monotonic Neural Activation Function
+    - https://arxiv.org/abs/1908.08681
+  """
+  return x.mul(F.softplus(x).tanh())
+
+
+class Mish(nn.Module):
+  """Mish: A Self Regularized Non-Monotonic Neural Activation Function
+    - https://arxiv.org/abs/1908.08681
+  """
+
+  def __init__(self, inplace: bool = False):
+    super(Mish, self).__init__()
+
+  def forward(self, x):
+    return mish(x)
+
+
+def sigmoid(x, inplace: bool = False):
+  return x.sigmoid_() if inplace else x.sigmoid()
