@@ -12,15 +12,19 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 # ==============================================================================
-r"""Losses collections
+"""Losses collections
 """
 
 import itertools
 import math
+
+import scipy.sparse as sp
 import numpy as np
+
 import torch
 from torch import nn
 import torch.nn.functional as F
+from torch.nn.modules.loss import _Loss
 
 # for autograd
 from torch.autograd.function import once_differentiable
@@ -32,6 +36,44 @@ except ImportError:
 
 import tw
 
+__all__ = [
+    "L1Loss",
+    "WeightedTVLoss",
+    "PerceptualLoss",
+    "LPIPSLoss",
+    "CharbonnierLoss",
+    "GeneralGanLoss",
+    "GradientPenaltyLoss",
+    "ContentLoss",
+    "AngleLoss",
+    "KLStandardGaussianLoss",
+    "LogRatioMetricLoss",
+    "PSNRLoss",
+    "PixelPositionAwareLoss",
+    "ReliabLoss",
+    "StructuralSimilarityLoss",
+    "SmoothL1Loss",
+    "IoULoss",
+    "GIoULoss",
+    "DIoULoss",
+    "CIoULoss",
+    "SigmoidFocalLoss",
+    "OrderSensitiveMetricLoss",
+    "MutualChannelLoss",
+    "LabelSmoothLoss",
+    "EBMLoss",
+    "MonotonicityRelatedLoss",
+    "PLCCLoss",
+    "LaplacianLoss",
+    "NMELoss",
+    "WingLoss",
+    "SmoothWingLoss",
+    "WiderWingLoss",
+    "NormalizedWiderWingLoss",
+    "MultiLevelEPELoss",
+    "MultiLevelCharbonnierLoss",
+]
+
 
 #!<-----------------------------------------------------------------------------
 #!<  spatial Loss
@@ -63,7 +105,7 @@ class L1Loss(nn.Module):
         weight (Tensor, optional): of shape (N, C, H, W). Element-wise
             weights. Default: None.
     """
-    return self.loss_weight * l1_loss(pred, target, weight, reduction=self.reduction)
+    return self.loss_weight * F.l1_loss(pred, target, weight, reduction=self.reduction)
 
 
 class WeightedTVLoss(L1Loss):
@@ -171,9 +213,11 @@ class PerceptualLoss(nn.Module):
       style_loss = 0
       for k in x_features.keys():
         if self.criterion_type == 'fro':
-          style_loss += torch.norm(self._gram_mat(x_features[k]) - self._gram_mat(gt_features[k]), p='fro') * self.layer_weights[k]
+          style_loss += torch.norm(self._gram_mat(x_features[k]) -
+                                   self._gram_mat(gt_features[k]), p='fro') * self.layer_weights[k]
         else:
-          style_loss += self.criterion(self._gram_mat(x_features[k]), self._gram_mat(gt_features[k])) * self.layer_weights[k]
+          style_loss += self.criterion(self._gram_mat(x_features[k]),
+                                       self._gram_mat(gt_features[k])) * self.layer_weights[k]
       style_loss *= self.style_weight
     else:
       style_loss = None
@@ -445,17 +489,17 @@ class AngleLoss(nn.Module):
     index = index.bool()
     index = torch.autograd.Variable(index)
 
-    self.lamb = max(self.LambdaMin, self.LambdaMax/(1+0.1*self.it))
+    self.lamb = max(self.LambdaMin, self.LambdaMax / (1 + 0.1 * self.it))
     output = cos_theta * 1.0  # size=(B,Classnum)
-    output[index] -= cos_theta[index]*(1.0+0)/(1+self.lamb)
-    output[index] += phi_theta[index]*(1.0+0)/(1+self.lamb)
+    output[index] -= cos_theta[index] * (1.0 + 0) / (1 + self.lamb)
+    output[index] += phi_theta[index] * (1.0 + 0) / (1 + self.lamb)
 
     logpt = F.log_softmax(output)
     logpt = logpt.gather(1, target)
     logpt = logpt.view(-1)
     pt = torch.autograd.Variable(logpt.data.exp())
 
-    loss = -1 * (1-pt)**self.gamma * logpt
+    loss = -1 * (1 - pt)**self.gamma * logpt
     loss = loss.mean()
 
     return loss
@@ -742,8 +786,8 @@ class StructuralSimilarityLoss(torch.nn.Module):
 
   @staticmethod
   def gaussian(window_size, sigma):
-    gauss = torch.Tensor([math.exp(-(x - window_size//2)**2/float(2*sigma**2)) for x in range(window_size)])  # nopep8
-    return gauss/gauss.sum()
+    gauss = torch.Tensor([math.exp(-(x - window_size // 2)**2 / float(2 * sigma**2)) for x in range(window_size)])  # nopep8
+    return gauss / gauss.sum()
 
   @staticmethod
   def create_window(window_size, channel):
@@ -754,21 +798,21 @@ class StructuralSimilarityLoss(torch.nn.Module):
 
   @staticmethod
   def ssim(img1, img2, window, window_size, channel, size_average=True):
-    mu1 = F.conv2d(img1, window, padding=window_size//2, groups=channel)
-    mu2 = F.conv2d(img2, window, padding=window_size//2, groups=channel)
+    mu1 = F.conv2d(img1, window, padding=window_size // 2, groups=channel)
+    mu2 = F.conv2d(img2, window, padding=window_size // 2, groups=channel)
 
     mu1_sq = mu1.pow(2)
     mu2_sq = mu2.pow(2)
-    mu1_mu2 = mu1*mu2
+    mu1_mu2 = mu1 * mu2
 
-    sigma1_sq = F.conv2d(img1*img1, window, padding=window_size//2, groups=channel) - mu1_sq  # nopep8
-    sigma2_sq = F.conv2d(img2*img2, window, padding=window_size//2, groups=channel) - mu2_sq  # nopep8
-    sigma12 = F.conv2d(img1*img2, window, padding=window_size // 2, groups=channel) - mu1_mu2  # nopep8
+    sigma1_sq = F.conv2d(img1 * img1, window, padding=window_size // 2, groups=channel) - mu1_sq  # nopep8
+    sigma2_sq = F.conv2d(img2 * img2, window, padding=window_size // 2, groups=channel) - mu2_sq  # nopep8
+    sigma12 = F.conv2d(img1 * img2, window, padding=window_size // 2, groups=channel) - mu1_mu2  # nopep8
 
     C1 = 0.01**2
     C2 = 0.03**2
 
-    ssim_map = ((2*mu1_mu2 + C1)*(2*sigma12 + C2)) / ((mu1_sq + mu2_sq + C1)*(sigma1_sq + sigma2_sq + C2))  # nopep8
+    ssim_map = ((2 * mu1_mu2 + C1) * (2 * sigma12 + C2)) / ((mu1_sq + mu2_sq + C1) * (sigma1_sq + sigma2_sq + C2))  # nopep8
 
     if size_average:
       return ssim_map.mean()
@@ -824,7 +868,7 @@ class SmoothL1Loss(nn.Module):
     beta parameter.
   """
 
-  def __init__(self, beta=1./9, reduction=None):
+  def __init__(self, beta=1. / 9, reduction=None):
     super(SmoothL1Loss, self).__init__()
     self.beta = beta
     self.reduction = reduction
@@ -835,6 +879,106 @@ class SmoothL1Loss(nn.Module):
     loss = torch.where(cond, 0.5 * n ** 2 / self.beta, n - 0.5 * self.beta)
     return loss
 
+#!<-----------------------------------------------------------------------------
+#!< IoU Losses
+#!<-----------------------------------------------------------------------------
+
+
+class IoULoss(nn.Module):
+
+  def __init__(self, method='linear', reduction=None):
+    super(IoULoss, self).__init__()
+    self.reduction = reduction
+    self.method = method
+    assert self.method in ['linear', 'log']
+    assert self.reduction in ['none', 'mean', 'sum', None]
+
+  def forward(self, preds, targets, **kwargs):
+    """Computing the IoU loss between a set of predicted bboxes and target bboxes.
+      The loss is calculated as negative log of IoU.
+
+    Args:
+        preds ([torch.Tensor]): [N, 4] (x1, y1, x2, y2)
+        targets ([torch.Tensor]): [N, 4] (x1, y1, x2, y2)
+
+    Returns:
+        [losses]: [N, ]
+    """
+    ious = tw.transform.bbox.aligned_iou(preds, targets, mode='iou').clamp(min=1e-6)
+    if self.method == 'linear':
+      loss = 1 - ious
+    elif self.method == 'log':
+      loss = -ious.log()
+    else:
+      raise NotImplementedError
+    return loss
+
+
+class GIoULoss(nn.Module):
+
+  def __init__(self, reduction=None):
+    super(GIoULoss, self).__init__()
+    self.reduction = reduction
+
+  def forward(self, preds, targets, **kwargs):
+    """Generalized Intersection over Union: A Metric and A Loss for Bounding
+      Box Regression <https://arxiv.org/abs/1902.09630>
+
+    Args:
+        preds ([torch.Tensor]): [N, 4] (x1, y1, x2, y2)
+        targets ([torch.Tensor]): [N, 4] (x1, y1, x2, y2)
+
+    Returns:
+        [losses]: [N, ]
+    """
+    gious = tw.transform.bbox.aligned_iou(preds, targets, mode='giou')
+    loss = 1 - gious
+    return loss
+
+
+class DIoULoss(nn.Module):
+
+  def __init__(self, reduction=None):
+    super(DIoULoss, self).__init__()
+    self.reduction = reduction
+
+  def forward(self, preds, targets, **kwargs):
+    """Implementation of Distance-IoU Loss: Faster and Better
+      Learning for Bounding Box Regression, https://arxiv.org/abs/1911.08287`_.
+
+      Code is modified from https://github.com/Zzh-tju/DIoU.
+
+    Args:
+        preds ([torch.Tensor]): [N, 4] (x1, y1, x2, y2)
+        targets ([torch.Tensor]): [N, 4] (x1, y1, x2, y2)
+
+    Returns:
+        [losses]: [N, ]
+    """
+    dious = tw.transform.bbox.aligned_iou(preds, targets, mode='diou')
+    loss = 1 - dious
+    return loss
+
+
+class CIoULoss(nn.Module):
+
+  def __init__(self, reduction=None):
+    super(CIoULoss, self).__init__()
+    self.reduction = reduction
+
+  def forward(self, preds, targets, **kwargs):
+    """
+
+    Args:
+        preds ([torch.Tensor]): [N, 4]
+        targets ([torch.Tensor]): [N, 4]
+
+    Returns:
+        [losses]: [N, ]
+    """
+    cious = tw.transform.bbox.aligned_iou(preds, targets, mode='ciou')
+    loss = 1 - cious
+    return loss
 
 #!<-----------------------------------------------------------------------------
 #!<
@@ -872,7 +1016,7 @@ def sigmoid_focal_loss_cpu(logits, targets, gamma, alpha):
   alpha = alpha[0]
   dtype = targets.dtype
   device = targets.device
-  class_range = torch.arange(1, num_classes+1, dtype=dtype, device=device).unsqueeze(0)
+  class_range = torch.arange(1, num_classes + 1, dtype=dtype, device=device).unsqueeze(0)
   t = targets.unsqueeze(1)
   p = torch.sigmoid(logits)
   term1 = (1 - p) ** gamma * torch.log(p)
@@ -1205,8 +1349,8 @@ class LabelSmoothLoss(nn.modules.loss._WeightedLoss):
     assert 0 <= smoothing < 1
     with torch.no_grad():
       targets = torch.empty(size=(targets.size(0), n_classes), device=targets.device) \
-          .fill_(smoothing / (n_classes-1)) \
-          .scatter_(1, targets.data.unsqueeze(1), 1.-smoothing)
+          .fill_(smoothing / (n_classes - 1)) \
+          .scatter_(1, targets.data.unsqueeze(1), 1. - smoothing)
     return targets
 
   def forward(self, inputs, targets):
@@ -1244,12 +1388,9 @@ class LabelSmoothLoss(nn.modules.loss._WeightedLoss):
 class EBMLoss(nn.Module):
   r"""Energy Based Loss"""
 
-  def __init__(self, num_samples,
-               gmm_stds,
-               method='nce',
-               **kwargs):
+  def __init__(self, num_samples, gmm_stds, method='nce', **kwargs):
     super(EBMLoss, self).__init__()
-    from tw.nn.random import GMMSampler
+    from tw.nn import GMMSampler
 
     self.num_samples = num_samples
     self.sampler = GMMSampler(gmm_stds)
@@ -1277,8 +1418,8 @@ class EBMLoss(nn.Module):
     else:
       raise NotImplementedError(method)
 
-  def forward(self, predictor, features, labels, device):
-    r"""Eenergy-based Loss
+  def forward(self, predictor, features, labels):
+    r"""Energy-based Loss
 
       Args:
         features: [N, C]
@@ -1298,9 +1439,9 @@ class EBMLoss(nn.Module):
     assert features.dim() == 2
     assert len(labels.shape) == 1
 
-    labels = labels.float().unsqueeze(1).to(device)
+    labels = labels.float().unsqueeze(1).to(features.device)
 
-    return self.ebm_loss(predictor, features, labels, device)
+    return self.ebm_loss(predictor, features, labels, features.device)
 
   def mcis(self, predictor, features, labels, device):
     r"""Monte-Carlo Importance Sampler: Negative Log-Likehood Minimize"""
@@ -1412,3 +1553,650 @@ class EBMLoss(nn.Module):
     loss = torch.mean(torch.mean(scores, dim=1) - scores_gt)
 
     return loss
+
+#!<-----------------------------------------------------------------------------
+#!< Related Loss
+#!< https://github.com/lidq92/MDTVSFA/blob/main/VQAloss.py
+#!<-----------------------------------------------------------------------------
+
+
+class MonotonicityRelatedLoss(nn.Module):
+
+  def __init__(self):
+    super(MonotonicityRelatedLoss, self).__init__()
+
+  def forward(self, inputs, targets):
+    """L = max{(y_i - y_j) * sign(y_j' - y_i'), 0}
+      it will take effects only for the un-consistent between prediction order
+      and groundtruth order.
+
+    Args:
+        inputs ([N, C]): inputs value
+        targets ([N, C]): targets value
+    """
+    bs = inputs.size(0)
+    assert inputs.ndim == targets.ndim == 2
+    assert bs > 1, "at least input number larger than 2."
+    loss = torch.sum(F.relu((inputs - inputs.t()) * torch.sign(targets.t() - targets)))
+    return loss / bs / (bs - 1)
+
+
+class PLCCLoss(nn.Module):
+
+  def __init__(self):
+    super(PLCCLoss, self).__init__()
+
+  def forward(self, inputs, targets):
+    """PLCC = \frac{(y-y_mean)(y'-y'_mean)}{(y-y_mean)^2(y'-y'_mean)^2}
+      it could be seen as cosine_similarity metric.
+
+    Args:
+        inputs ([N, C]): inputs value
+        targets ([N, C]): targets value
+    """
+    bs = inputs.size(0)
+    assert inputs.ndim == targets.ndim == 2
+    assert bs > 1, "at least input number larger than 2."
+    return (1 - torch.cosine_similarity(inputs.t() - torch.mean(inputs), targets.t() - torch.mean(targets))[0]) / 2
+
+#!<-----------------------------------------------------------------------------
+#!< Face alignment related loss
+#!< https://github.com/BeierZhu/Sturcture-Coherency-Face-Alignment
+#!<-----------------------------------------------------------------------------
+
+
+def gen_laplacian_matrix(num_points, top_k):
+  """
+  Return 106 x 106 matrix
+  """
+  if num_points == 98:
+    connection = np.loadtxt(f'utils/face_graph/{num_points}pts_top{top_k}.txt')
+  else:
+    connection = np.loadtxt(f'utils/face_graph/{num_points}pts.txt')
+
+  return adj_mx_from_list(connection, num_points)
+
+
+def adj_mx_from_list(connection, num_points):
+  edges = connection[:, :2]
+  return adj_mx_from_weighted_edges(num_points, edges)
+
+
+def adj_mx_from_weighted_edges(num_pts, edges):
+  edges = np.array(edges, dtype=np.int32)
+  weights = np.ones(edges.shape[0])
+  data, i, j = weights, edges[:, 0], edges[:, 1]
+  adj_mx = sp.coo_matrix((data, (i, j)), shape=(num_pts, num_pts), dtype=np.float32).todense()
+
+  np.fill_diagonal(adj_mx, 0)
+  row_sum = adj_mx.sum(axis=1)
+  row_sum[row_sum == 0] = 1
+  adj_mx /= row_sum
+  np.fill_diagonal(adj_mx, -1)
+
+  adj_mx = torch.tensor(adj_mx, dtype=torch.float)
+  return adj_mx.unsqueeze(0)
+
+
+class LaplacianLoss(_Loss):
+  def __init__(self, loss, num_points, reduction="mean"):
+    super(LaplacianLoss, self).__init__(reduction=reduction)
+    self.loss = loss
+    self.num_points = num_points
+    self.laplacian_matrix = gen_laplacian_matrix(num_points, 3).cuda()  # 106x106
+
+  def forward(self, prediction, target):
+    prediction = prediction.reshape(-1, self.num_points, 2)
+    target = target.reshape(-1, self.num_points, 2)
+    prediction = torch.matmul(self.laplacian_matrix, prediction)
+    target = torch.matmul(self.laplacian_matrix, target)
+    prediction = prediction.reshape(-1, self.num_points*2)
+    target = target.reshape(-1, self.num_points*2)
+
+    return self.loss(prediction, target)
+
+
+def nme_loss(prediction: torch.Tensor, target: torch.Tensor, num_points, image_size=256, reduction="mean"):
+  """
+  normalized_wider_wing = 
+      1. C/d * |x|                           , if |x| < thres1
+      2. C/d * (thres2*ln(1+|x|/curvature) + B), if |x| >= thres2
+  """
+  assert prediction.size() == target.size(), 'input tensors must have the same size'
+
+  # Calculate Euclidean distances between actual and target locations
+  diff = prediction - target
+  diff = diff.reshape(diff.size(0), -1, 2)
+  dist_sq = diff.pow(2).sum(-1, keepdim=False)
+  loss = dist_sq.sqrt()  # shape (B x N)
+
+  if num_points == 29:
+    le_idx = [8]
+    re_idx = [9]
+
+  if num_points == 68:
+    le_idx = [36]  # le = left eye
+    re_idx = [45]  # re = right eye
+
+  if num_points == 98:
+    le_idx = [60]
+    re_idx = [72]
+
+  target = target.reshape(-1, num_points, 2)
+  le_loc = torch.mean(target[:, le_idx, :], dim=1)  # batchsize x 2
+  re_loc = torch.mean(target[:, re_idx, :], dim=1)
+
+  norm_dist = torch.sqrt(torch.sum((le_loc - re_loc)**2, dim=1))  # batchsize
+  factor = image_size/norm_dist
+
+  if reduction == "sum":
+    loss = loss.sum(dim=1)
+    loss *= factor
+    loss = loss.sum()
+
+  if reduction == "mean":
+    loss = loss.mean(dim=1)
+    loss *= factor
+    loss = loss.mean()
+
+  return loss
+
+
+class NMELoss(_Loss):
+  """docstring for NMELoss"""
+
+  def __init__(self, num_points, image_size, reduction="mean"):
+    super(NMELoss, self).__init__(reduction=reduction)
+    self.num_points = num_points
+    self.image_size = image_size
+
+  def forward(self, prediction, target):
+    return nme_loss(prediction, target,
+                    num_points=self.num_points,
+                    image_size=self.image_size,
+                    reduction=self.reduction)
+
+
+def wing_loss(prediction: torch.Tensor, target: torch.Tensor, width=5, curvature=0.5, reduction="mean"):
+  """
+  https://arxiv.org/pdf/1711.06753.pdf
+  :param prediction:
+  :param target:
+  :param width:
+  :param curvature:
+  :param reduction:
+  :return:
+  """
+  diff_abs = (target - prediction).abs()
+  loss = diff_abs.clone()
+
+  idx_smaller = diff_abs < width
+  idx_bigger = diff_abs >= width
+
+  loss[idx_smaller] = width * torch.log(1 + diff_abs[idx_smaller] / curvature)
+
+  C = width - width * math.log(1 + width / curvature)
+  loss[idx_bigger] = loss[idx_bigger] - C
+
+  if reduction == "sum":
+    loss = loss.sum()
+
+  if reduction == "mean":
+    loss = loss.mean()
+
+  return loss
+
+
+class WingLoss(_Loss):
+  def __init__(self, width=5, curvature=0.5, reduction="mean"):
+    super(WingLoss, self).__init__(reduction=reduction)
+    self.width = width
+    self.curvature = curvature
+
+  def forward(self, prediction, target):
+    return wing_loss(prediction, target, self.width, self.curvature, self.reduction)
+
+
+def smooth_wing_loss(prediction: torch.Tensor, target: torch.Tensor, thres1=2, thres2=20, curvature=2, reduction="mean"):
+  """
+  smooth_wing = 
+      1. |x|                           , if |x| < thres1
+      2. thres2*ln(1+|x|/curvature) + B, if A <= |x| < thres2
+      3. |x| - C                       , if |x| >= thres2
+  """
+  loss = (target - prediction).abs()
+
+  idx_small = loss < thres1
+  idx_normal = (loss >= thres1) * (loss < thres2)
+  idx_big = loss >= thres2
+
+  B = thres1 - thres2 * math.log(1 + thres1 / curvature)
+  loss[idx_normal] = thres2 * torch.log(1 + loss[idx_normal] / curvature) + B
+
+  C = thres2 - thres2 * math.log(1 + thres2 / curvature) - B
+  loss[idx_big] = loss[idx_big] - C
+
+  if reduction == "sum":
+    loss = loss.sum()
+  if reduction == "mean":
+    loss = loss.mean()
+
+  return loss
+
+
+class SmoothWingLoss(_Loss):
+  """docstring for SmoothWingLoss"""
+
+  def __init__(self, thres1=2, thres2=20, curvature=2, reduction="mean"):
+    super(SmoothWingLoss, self).__init__(reduction=reduction)
+    self.thres1 = thres1
+    self.thres2 = thres2
+    self.curvature = curvature
+
+  def forward(self, prediction, target):
+    return smooth_wing_loss(prediction, target, self.thres1, self.thres2, self.curvature, self.reduction)
+
+
+def wider_wing_loss(prediction: torch.Tensor, target: torch.Tensor, thres1=2, thres2=20, curvature=2, reduction="mean"):
+  """
+  wider_wing = 
+      1. |x|                           , if |x| < thres1
+      2. thres2*ln(1+|x|/curvature) + B, if |x| >= thres2
+  """
+  loss = (target - prediction).abs()
+
+  idx_small = loss < thres1
+  idx_big = loss >= thres1
+
+  B = thres1 - thres2 * math.log(1 + thres1 / curvature)
+  loss[idx_big] = thres2 * torch.log(1 + loss[idx_big] / curvature) + B
+
+  if reduction == "sum":
+    loss = loss.sum()
+  if reduction == "mean":
+    loss = loss.mean()
+
+  return loss
+
+
+class WiderWingLoss(_Loss):
+  """docstring for WiderWingLoss"""
+
+  def __init__(self, thres1=2, thres2=20, curvature=2, reduction="mean"):
+    super(WiderWingLoss, self).__init__(reduction=reduction)
+    self.thres1 = thres1
+    self.thres2 = thres2
+    self.curvature = curvature
+
+  def forward(self, prediction, target):
+    return wider_wing_loss(prediction, target, self.thres1, self.thres2, self.curvature, self.reduction)
+
+
+def normalized_wider_wing_loss(prediction: torch.Tensor, target: torch.Tensor, num_points, image_size=256, thres1=2, thres2=20, curvature=2, reduction="mean"):
+  """
+  normalized_wider_wing = 
+      1. C/d * |x|                           , if |x| < thres1
+      2. C/d * (thres2*ln(1+|x|/curvature) + B), if |x| >= thres2
+  """
+  loss = (target - prediction).abs()
+
+  idx_small = loss < thres1
+  idx_big = loss >= thres1
+
+  B = thres1 - thres2 * math.log(1 + thres1 / curvature)
+  loss[idx_big] = thres2 * torch.log(1 + loss[idx_big] / curvature) + B
+
+  if num_points == 29:
+    le_idx = [8]
+    re_idx = [9]
+
+  if num_points == 68:
+    le_idx = [36]  # le = left eye
+    re_idx = [45]  # re = right eye
+
+  if num_points == 98:
+    le_idx = [60]
+    re_idx = [72]
+
+  target = target.reshape(-1, num_points, 2)
+  le_loc = torch.mean(target[:, le_idx, :], dim=1)  # batchsize x 2
+  re_loc = torch.mean(target[:, re_idx, :], dim=1)
+
+  norm_dist = torch.sqrt(torch.sum((le_loc - re_loc)**2, dim=1))  # batchsize
+  factor = image_size/norm_dist
+
+  if reduction == "sum":
+    loss = loss.sum(dim=1)
+    loss *= factor
+    loss = loss.sum()
+
+  if reduction == "mean":
+    loss = loss.mean(dim=1)
+    loss *= factor
+    loss = loss.mean()
+
+  return loss
+
+
+class NormalizedWiderWingLoss(_Loss):
+  """docstring for NormalizedWiderWingLoss"""
+
+  def __init__(self, thres1=2, thres2=20, curvature=2, num_points=98, image_size=256, reduction="mean"):
+    super(NormalizedWiderWingLoss, self).__init__(reduction=reduction)
+    self.thres1 = thres1
+    self.thres2 = thres2
+    self.curvature = curvature
+
+    self.num_points = num_points
+    self.image_size = image_size
+
+  def forward(self, prediction, target):
+    return normalized_wider_wing_loss(prediction, target, self.num_points, self.image_size, self.thres1, self.thres2, self.curvature, self.reduction)
+
+#!<-----------------------------------------------------------------------------
+#!< EPE error: optical flow
+#!<-----------------------------------------------------------------------------
+
+
+def endpoint_error(pred, target, p=2, q=None, eps=None):
+  """Calculate end point errors between prediction and ground truth.
+
+  If not define q, the loss function is
+  .. math::
+    loss = \Vert \mathbf{u}-\mathbf{u_gt} \Vert^p
+
+  otherwise,
+  .. math::
+    loss = (\Vert \mathbf{u}-\mathbf{u_gt} \Vert^p+eps)^q
+
+  For PWC-Net L2 norm loss: p=2, for the robust loss function p=1, q=0.4,
+  eps=0.01.
+
+  Args:
+      pred (torch.Tensor): output flow map from flow_estimator
+          shape(B, 2, H, W).
+      target (torch.Tensor): ground truth flow map shape(B, 2, H, W).
+      p (int): norm degree for loss. Options are 1 or 2. Defaults to 2.
+      q (float, optional): used to give less penalty to outliers when
+          fine-tuning model. Defaults to 0.4.
+      eps (float, optional): a small constant to numerical stability when
+          fine-tuning model. Defaults to 0.01.
+
+  Returns:
+      Tensor: end-point error map with the shape (B, H, W).
+  """
+
+  assert pred.shape == target.shape, (f'pred shape {pred.shape} does not match target shape {target.shape}.')
+
+  epe_map = torch.norm(pred - target, p, dim=1)  # shape (B, H, W).
+
+  if q is not None and eps is not None:
+    epe_map = (epe_map + eps)**q
+
+  return epe_map
+
+
+class MultiLevelEPELoss(nn.Module):
+  """Multi-level end point error loss.
+
+  Args:
+
+    p (int): norm degree for loss. Options are 1 or 2. Defaults to 2.
+    q (float): used to give less penalty to outliers when fine-tuning model. Defaults to None.
+    eps (float): a small constant to numerical stability when fine-tuning model. Defaults to None.
+    weights (dict): manual rescaling weights given to the loss of flow map
+        at each level, and the keys in weights must correspond to predicted
+        dict. Defaults to dict(level6=0.32, level5=0.08, level4=0.02, level3=0.01, level2=0.005).
+    flow_div (float): the divisor used to scale down ground truth. Defaults to 20.
+    max_flow (float): maximum value of optical flow, if some pixel's flow
+        of target is larger than it, this pixel is not valid. Default to float('inf').
+    resize_flow (str): mode for reszing flow: 'downsample' and 'upsample',
+        as multi-level predicted outputs don't match the ground truth.
+        If set to 'downsample', it will downsample the ground truth, and
+        if set to 'upsample' it will upsample the predicted flow, and
+        'upsample' is used for sparse flow map as no generic interpolation
+        mode can resize a ground truth of sparse flow correctly. Default to 'downsample'.
+    scale_as_level (bool): Whether flow for each level is at its native
+        spatial resolution. If `'scale_as_level'` is True, the ground
+        truth is scaled at different levels, if it is False, the ground truth will not be scaled. Default to False.
+    reduction (str): the reduction to apply to the output:'none', 'mean',
+        'sum'. 'none': no reduction will be applied and will return a
+        full-size epe map, 'mean': the mean of the epe map is taken, 'sum':
+        the epe map will be summed but averaged by batch_size. Default: 'sum'.
+
+    """
+
+  def __init__(self, p=2,
+               q=None,
+               eps=None,
+               weights=[0.32, 0.08, 0.02, 0.01, 0.005],
+               flow_div=20.,
+               max_flow=float('inf'),
+               resize_flow='downsample',
+               scale_as_level=False,
+               reduction='sum'):
+    super(MultiLevelEPELoss, self).__init__()
+    assert p == 1 or p == 2
+    self.p = p
+    self.q = q
+    if self.q is not None:
+      assert self.q > 0
+
+    self.eps = eps
+    if self.eps is not None:
+      assert eps > 0
+
+    assert flow_div > 0
+    self.flow_div = flow_div
+
+    assert isinstance(weights, list)
+    self.weights = weights
+
+    assert max_flow > 0.
+    self.max_flow = max_flow
+
+    assert resize_flow in ('downsample', 'upsample')
+    self.resize_flow = resize_flow
+
+    assert isinstance(scale_as_level, bool)
+    self.scale_as_level = scale_as_level
+
+    assert reduction in ('mean', 'sum')
+    self.reduction = reduction
+
+  def forward(self, pred_list, target, valid=None):
+    """Forwar function for MultiLevelCharbonnierLoss.
+
+    Args:
+        pred_list (list): Multi-level output of predicted optical flow,
+            and the contain in dict is a Tensor or list of Tensor with
+            shape (B, 2, H_l, W_l), where l indicates the level.
+        target (Tensor): Ground truth of optical flow with shape (B, 2, H, W).
+        valid (Tensor, optional): Valid mask for optical flow. Defaults to None.
+
+    Returns:
+        Tensor: value of pixel-wise generalized Charbonnier loss.
+
+    """
+    # compute valid optical flow (avoid to outlier)
+    if valid is None:
+      valid = torch.ones_like(target[:, 0, :, :])
+    else:
+      mag = torch.sum(target**2, dim=1).sqrt()
+      valid = ((valid >= 0.5) & (mag < self.max_flow)).to(target)
+
+    # normalize
+    th, tw = target.shape[2:]
+    target_div = target / self.flow_div
+
+    # optical flow target shape
+    c_org, h_org, w_org = target.shape[1:]
+    assert c_org == 2, f'The channels ground truth must be 2, but got {c_org}'
+
+    # over for each level: from top to bottom
+    loss = 0
+    for level, cur_weight in enumerate(self.weights):
+      cur_pred = pred_list[level] / self.flow_div
+      b, _, h, w = cur_pred.shape
+
+      if self.resize_flow == 'downsample':
+        # down sample ground truth following irr solution
+        # https://github.com/visinf/irr/blob/master/losses.py#L16
+        cur_target = F.adaptive_avg_pool2d(target_div, [h, w])
+        if self.scale_as_level:
+          cur_target[:, 0] /= (tw / w)
+          cur_target[:, 1] /= (th / h)
+        cur_valid = F.adaptive_max_pool2d(valid, [h, w])
+      else:
+        cur_target = target_div
+        cur_valid = valid
+
+      if self.resize_flow == 'upsample':
+        # up sample predicted flow following pwcnet and irr solution
+        # https://github.com/visinf/irr/blob/master/losses.py#L20
+        # when training sparse flow dataset, as no generic
+        # interpolation mode can resize a ground truth of sparse flow correctly.
+        cur_pred = F.interpolate(cur_pred, size=cur_target.shape[2:], mode='bilinear', align_corners=False)
+        if self.scale_as_level:
+          cur_pred[:, 0] *= (tw / w)
+          cur_pred[:, 1] *= (th / h)
+
+      loss_map = endpoint_error(cur_pred, cur_target, p=self.p, q=self.q, eps=self.eps) * cur_valid
+
+      if self.reduction == 'mean':
+        loss += loss_map.sum() / (cur_valid.sum() + 1e-8) * cur_weight
+      elif self.reduction == 'sum':
+        loss += loss_map.sum() / b * cur_weight
+
+    return loss
+
+
+def charbonnier_loss(pred, target, q=0.2, eps=0.01):
+  """Generalized Charbonnier loss function between output and ground truth.
+
+  The loss function is
+  .. math::
+    loss = ((u-u_gt)^2+(v-v_gt)^2+eps)^q
+
+  Generalized Charbonnier loss was used in LiteFlowNet when fine tuning,
+  with eps=0.01 q=0.2.
+
+  Args:
+      pred (torch.Tensor): output flow map from flow_estimator
+          shape(B, 2, H, W).
+      target (torch.Tensor): ground truth flow map shape(B, 2, H, W).
+      q (float): the exponent in charbonnier loss.
+      eps (float): small constant to numerical stability when
+          fine-tuning model. Defaults to 0.01.
+
+  Returns:
+      Tensor: loss map with the shape (B, H, W).
+  """
+
+  assert pred.shape == target.shape, (f'pred shape {pred.shape} does not match target shape {target.shape}.')
+  diff = torch.add(pred, -target)
+  loss_map = (torch.sum(diff * diff, dim=1) + eps)**q  # shape (B, H, W).
+  return loss_map
+
+
+class MultiLevelCharbonnierLoss(nn.Module):
+
+  def __init__(self, q=0.2,
+               eps=0.01,
+               flow_div=20.,
+               weights=[0.32, 0.08, 0.02, 0.01, 0.005],
+               max_flow=float('inf'),
+               resize_flow='downsample',
+               scale_as_level=False,
+               reduction='sum'):
+    super(MultiLevelCharbonnierLoss, self).__init__()
+
+    assert isinstance(q, float) and q > 0.
+    self.q = q
+
+    assert isinstance(eps, float) and eps > 0.
+    self.eps = eps
+
+    assert flow_div > 0
+    self.flow_div = flow_div
+
+    assert isinstance(weights, list)
+    self.weights = weights
+
+    assert max_flow > 0.
+    self.max_flow = max_flow
+
+    assert resize_flow in ('downsample', 'upsample')
+    self.resize_flow = resize_flow
+
+    assert isinstance(scale_as_level, bool)
+    self.scale_as_level = scale_as_level
+
+    assert reduction in ('mean', 'sum')
+    self.reduction = reduction
+
+  def forward(self, pred_list, target, valid=None):
+    """Forwar function for MultiLevelCharbonnierLoss.
+
+    Args:
+        preds_list (list): Multi-level output of predicted optical flow,
+            and the contain in dict is a Tensor or list of Tensor with
+            shape (B, 2, H_l, W_l), where l indicates the level.
+        target (Tensor): Ground truth of optical flow with shape (B, 2, H, W).
+        valid (Tensor, optional): Valid mask for optical flow. Defaults to None.
+
+    Returns:
+        Tensor: value of pixel-wise generalized Charbonnier loss.
+
+    """
+    # compute valid optical flow (avoid to outlier)
+    if valid is None:
+      valid = torch.ones_like(target[:, 0, :, :])
+    else:
+      mag = torch.sum(target**2, dim=1).sqrt()
+      valid = ((valid >= 0.5) & (mag < self.max_flow)).to(target)
+
+    # normalize
+    th, tw = target.shape[2:]
+    target_div = target / self.flow_div
+
+    # optical flow target shape
+    c_org, h_org, w_org = target.shape[1:]
+    assert c_org == 2, f'The channels ground truth must be 2, but got {c_org}'
+
+    # over for each level: from top to bottom
+    loss = 0
+    for level, cur_weight in enumerate(self.weights):
+      cur_pred = pred_list[level] / self.flow_div
+      b, _, h, w = cur_pred.shape
+
+      if self.resize_flow == 'downsample':
+        # down sample ground truth following irr solution
+        # https://github.com/visinf/irr/blob/master/losses.py#L16
+        cur_target = F.adaptive_avg_pool2d(target_div, [h, w])
+        if self.scale_as_level:
+          cur_target[:, 0] /= (tw / w)
+          cur_target[:, 1] /= (th / h)
+        cur_valid = F.adaptive_max_pool2d(valid, [h, w])
+      else:
+        cur_target = target_div
+        cur_valid = valid
+
+      if self.resize_flow == 'upsample':
+        # up sample predicted flow following pwcnet and irr solution
+        # https://github.com/visinf/irr/blob/master/losses.py#L20
+        # when training sparse flow dataset, as no generic
+        # interpolation mode can resize a ground truth of sparse flow correctly.
+        cur_pred = F.interpolate(cur_pred, size=cur_target.shape[2:], mode='bilinear', align_corners=False)
+        if self.scale_as_level:
+          cur_pred[:, 0] *= (tw / w)
+          cur_pred[:, 1] *= (th / h)
+
+      loss_map = charbonnier_loss(cur_pred, cur_target, q=self.q, eps=self.eps) * cur_valid
+
+      if self.reduction == 'mean':
+        loss += loss_map.sum() / (cur_valid.sum() + 1e-8) * cur_weight
+      elif self.reduction == 'sum':
+        loss += loss_map.sum() / b * cur_weight
+
+    return loss
