@@ -20,6 +20,7 @@ from torch import nn
 from torch.nn import functional as F
 from torch.autograd import Function
 from torch.autograd.function import once_differentiable
+from torch.nn.modules.pooling import AdaptiveMaxPool2d
 from torch.nn.modules.utils import _pair
 
 try:
@@ -27,6 +28,49 @@ try:
 except ImportError:
   _C = None
 
+__all__ = [
+    'DeformRoIPoolingFunction',
+    'DeformRoIPooling',
+    'DeformRoIPoolingPack',
+    'ModulatedDeformRoIPoolingPack',
+    'RoIAlign',
+    'RoIPool',
+    'ChannelMaxPool',
+    'ChannelAvgPool',
+    'CrissCrossAttention',
+    'AtrousSpatialPyramidPooling',
+    'AdaptiveAvgMaxPool2d',
+    'AdaptiveCatAvgMaxPool2d',
+]
+
+#!<-----------------------------------------------------------------------------
+#!< Pooling
+#!<-----------------------------------------------------------------------------
+
+
+class AdaptiveAvgMaxPool2d(nn.Module):
+
+  def ___init__(self, output_size=1):
+    super(AdaptiveAvgMaxPool2d, self).__init__()
+    self.output_size = output_size
+
+  def forward(self, x):
+    x_avg = F.adaptive_avg_pool2d(x, self.output_size)
+    x_max = F.adaptive_max_pool2d(x, self.output_size)
+    return 0.5 * (x_avg + x_max)
+
+
+class AdaptiveCatAvgMaxPool2d(nn.Module):
+
+  def ___init__(self, output_size=1):
+    super(AdaptiveCatAvgMaxPool2d, self).__init__()
+    self.output_size = output_size
+
+  def forward(self, x):
+    x_avg = F.adaptive_avg_pool2d(x, self.output_size)
+    x_max = F.adaptive_max_pool2d(x, self.output_size)
+    return torch.cat((x_avg, x_max), 1)
+
 #!<-----------------------------------------------------------------------------
 #!< Deformable Pooling
 #!<-----------------------------------------------------------------------------
@@ -289,7 +333,8 @@ class _RoIAlign(Function):
     spatial_scale = ctx.spatial_scale
     sampling_ratio = ctx.sampling_ratio
     bs, ch, h, w = ctx.input_shape
-    grad_input = _C.roi_align_backward(grad_output, rois, spatial_scale, output_size[0], output_size[1], bs, ch, h, w, sampling_ratio)
+    grad_input = _C.roi_align_backward(grad_output, rois, spatial_scale,
+                                       output_size[0], output_size[1], bs, ch, h, w, sampling_ratio)
     return grad_input, None, None, None, None
 
 
@@ -329,18 +374,45 @@ class _ROIPool(Function):
     output_size = ctx.output_size
     spatial_scale = ctx.spatial_scale
     bs, ch, h, w = ctx.input_shape
-    grad_input = _C.roi_pool_backward(grad_output, inputs, rois, argmax, spatial_scale, output_size[0], output_size[1], bs, ch, h, w)
+    grad_input = _C.roi_pool_backward(grad_output, inputs, rois, argmax, spatial_scale,
+                                      output_size[0], output_size[1], bs, ch, h, w)
     return grad_input, None, None, None
 
 
-class ROIPool(nn.Module):
+class RoIPool(nn.Module):
   def __init__(self, output_size, spatial_scale):
-    super(ROIPool, self).__init__()
+    """Performs Region of Interest (RoI) Pool operator described in Fast R-CNN
+
+    Args:
+        output_size ([int]): the size of the output after the cropping is performed, as (height, width)
+        spatial_scale ([float]): a scaling factor that maps the input coordinates to the box coordinates.
+        rois: (Tensor[K, 5] or List[Tensor[L, 4]]): the box coordinates in (x1, y1, x2, y2)
+            format where the regions will be taken from.
+            The coordinate must satisfy ``0 <= x1 < x2`` and ``0 <= y1 < y2``.
+            If a single Tensor is passed, then the first column should
+            contain the index of the corresponding element in the batch, i.e. a number in ``[0, N - 1]``.
+            If a list of Tensors is passed, then each Tensor will correspond to the boxes for an element i
+            in the batch.
+
+    Returns:
+        Tensor[K, C, output_size[0], output_size[1]]
+
+    """
+    super(RoIPool, self).__init__()
     self.output_size = output_size
     self.spatial_scale = spatial_scale
 
   def forward(self, inputs, rois):
-    return _ROIPool.apply(input, rois, self.output_size, self.spatial_scale)
+    """roi max pooling.
+
+    Args:
+        inputs ([torch.Tensor]): [N, C, H, W]
+        rois ([torch.Tensor]): [K, 5] (batch_ind, x1, y1, x2, y2)
+
+    Returns:
+        [type]: [description]
+    """
+    return _ROIPool.apply(inputs, rois, self.output_size, self.spatial_scale)
 
   def __repr__(self):
     tmpstr = self.__class__.__name__ + "("
@@ -360,8 +432,8 @@ class ChannelMaxPool(nn.modules.pooling._MaxPoolNd):
   r"""Applies a channel max pooling over an input signal composed of several input
   planes.
 
-    In the simplest case, the output value of the layer with input size 
-  :math:`(N, C, H, W)` and output :math:`(N, C_{out}, H, W)` can be precisely 
+    In the simplest case, the output value of the layer with input size
+  :math:`(N, C, H, W)` and output :math:`(N, C_{out}, H, W)` can be precisely
   described as:
 
   """
@@ -392,8 +464,8 @@ class ChannelAvgPool(nn.modules.pooling._AvgPoolNd):
   r"""Applies a channel max pooling over an input signal composed of several input
   planes.
 
-    In the simplest case, the output value of the layer with input size 
-  :math:`(N, C, H, W)` and output :math:`(N, C_{out}, H, W)` can be precisely 
+    In the simplest case, the output value of the layer with input size
+  :math:`(N, C, H, W)` and output :math:`(N, C_{out}, H, W)` can be precisely
   described as:
 
   """
@@ -515,7 +587,7 @@ class AtrousSpatialPyramidPooling(nn.Module):
 
   DeepLab: Semantic Image Segmentation with Deep Convolutional Nets,
     Atrous Convolution, and Fully Connected CRFs. IEEE Transactions on
-    Pattern Analysis and Machine Intelligence, 40(4), 834–848.        
+    Pattern Analysis and Machine Intelligence, 40(4), 834–848.
     https://doi.org/10.1109/TPAMI.2017.2699184
 
   For backbone:
@@ -535,10 +607,14 @@ class AtrousSpatialPyramidPooling(nn.Module):
     else:
       raise NotImplementedError
 
-    self.aspp1 = _ASPPModule(in_channels, out_channels, kernel_size=1, padding=0, dilation=dilations[0], batchnorm=batchnorm)
-    self.aspp2 = _ASPPModule(in_channels, out_channels, kernel_size=3, padding=dilations[1], dilation=dilations[1], batchnorm=batchnorm)
-    self.aspp3 = _ASPPModule(in_channels, out_channels, kernel_size=3, padding=dilations[2], dilation=dilations[2], batchnorm=batchnorm)
-    self.aspp4 = _ASPPModule(in_channels, out_channels, kernel_size=3, padding=dilations[3], dilation=dilations[3], batchnorm=batchnorm)
+    self.aspp1 = _ASPPModule(in_channels, out_channels, kernel_size=1, padding=0,
+                             dilation=dilations[0], batchnorm=batchnorm)
+    self.aspp2 = _ASPPModule(in_channels, out_channels, kernel_size=3,
+                             padding=dilations[1], dilation=dilations[1], batchnorm=batchnorm)
+    self.aspp3 = _ASPPModule(in_channels, out_channels, kernel_size=3,
+                             padding=dilations[2], dilation=dilations[2], batchnorm=batchnorm)
+    self.aspp4 = _ASPPModule(in_channels, out_channels, kernel_size=3,
+                             padding=dilations[3], dilation=dilations[3], batchnorm=batchnorm)
 
     self.global_avg_pool = nn.Sequential(
         nn.AdaptiveAvgPool2d((1, 1)),
